﻿namespace Core.Module.HttpFilter
{
    using System;
    using System.Collections.Generic;
    using System.Reflection;
    using Core.AssemblyType;

    internal sealed class FilterTypeCache
    {
        private const string _typeCacheName = "Http-FilterTypeCache.xml";

        //缓存方法的委托，Key为方法的hashcode
        private Dictionary<int, Action<FilterContext>> _cacheDelMethod = new Dictionary<int, Action<FilterContext>>();
        //缓存控制器Action与指定方法的映射关系，其中Key为ControllerNameActionName
        private Dictionary<string, List<int>> _cacheMapper;
        private object _lockObj = new object();

        public void EnsureInitialized(IBuildManager buildManager)
        {
            if (_cacheMapper == null)
            {
                lock (_lockObj)
                {
                    _cacheMapper = new Dictionary<string, List<int>>();
                    List<Type> filterTypes = TypeCacheUtil.GetFilteredTypesFromAssemblies(_typeCacheName, IsFilterType, buildManager);
                    foreach (var t in filterTypes)
                    {
                        //该类型对象实例
                        object instance = null;
                        //所有公开、非继承的实例方法
                        //var methods = t.GetType().GetMethods(BindingFlags.Instance & BindingFlags.Public & BindingFlags.DeclaredOnly);
                        var methods = t.GetMethods(BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly);
                        //是否映射过该方法，用于是否添加方法委托至缓存
                        var anyMapper = false;
                        foreach (var m in methods)
                        {
                            //仅加载类型为Action<FilterContext>的方法
                            var paras = m.GetParameters();
                            if (paras.Length != 1 || paras[0].ParameterType != typeof(FilterContext))
                                continue;
                            //读取方法的Attribute
                            var mCode = m.GetHashCode();
                            var attrs = m.GetCustomAttributes(typeof(FilterMethodAttribute), false);
                            //组织Action与类方法的HashCode的映射关系
                            foreach (FilterMethodAttribute attr in attrs)
                            {
                                foreach (var action in attr.actionName)
                                {
                                    var key = string.Format("{0}{1}", attr.controllerName, action).ToLower();
                                    List<int> map;
                                    if (!_cacheMapper.TryGetValue(key, out map))
                                    {
                                        map = new List<int>() { mCode };
                                        _cacheMapper[key] = map;
                                    }
                                    else
                                    {
                                        map.Add(mCode);
                                    }
                                    if (!anyMapper) anyMapper = true;
                                }
                            }
                            //存储方法委托,忽略未映射过的方法
                            if (!anyMapper) continue;
                            if (instance == null)
                            { //仅当需要创建方法委托时才创建对象实例，已创建过则忽略
                                instance = Activator.CreateInstance(t);
                            }
                            var mDel = Delegate.CreateDelegate(typeof(Action<FilterContext>), instance, m) as Action<FilterContext>;
                            _cacheDelMethod.Add(mCode, mDel);
                        }
                    }

                }
            }
        }

        internal static bool IsFilterType(Type t)
        {
            return
                t != null &&
                t.IsPublic &&
                !t.IsAbstract &&
                typeof(IFilter).IsAssignableFrom(t);
        }

        public List<Action<FilterContext>> GetFilterAction(string controllerName, string actionName)
        {
            List<Action<FilterContext>> actionList = new List<Action<FilterContext>>();
            var key = string.Format("{0}{1}", controllerName, actionName).ToLower();
            var lstHashCode = new List<int>();
            if (_cacheMapper.TryGetValue(key, out lstHashCode))
            {
                foreach (var code in lstHashCode)
                {
                    actionList.Add(_cacheDelMethod[code]);
                }
            }
            return actionList;
        }

        public bool HasFilterAction(string controllerName, string actionName)
        {
            var key = string.Format("{0}{1}", controllerName, actionName).ToLower();
            var lstHashCode = new List<int>();
            if (_cacheMapper.TryGetValue(key, out lstHashCode))
            {
                return lstHashCode.Count > 0;
            }
            return false;
        }
    }
}